#include "LorentzCell.h"

// Constructor for LorentzCell Class
LorentzCell::LorentzCell(int numPoles, const double *plasmon, const double *oscillator, const double *damping){
	double sumGamma = 0.0; // An auxiliarry variable to hold intermediate results

	(*this).numPoles = numPoles;
	Exold = Eyold = Ezold = Exnew = Eynew = Eznew = 0.0;
	outJx = outJy = outJz = 0.0;
	
	Jx    = new double[numPoles];
	Jy    = new double[numPoles];
	Jz	  = new double[numPoles];
	Jxold = new double[numPoles];
	Jyold = new double[numPoles];
	Jzold = new double[numPoles];
	Jxnew = new double[numPoles];
	Jynew = new double[numPoles];
	Jznew = new double[numPoles];
	alpha = new double[numPoles];
	gamma = new double[numPoles];
	zeta  = new double[numPoles];
	
	for(int i=0;i<numPoles;i++){

		// Set the coefficient arrays
		alpha[i] = (2-delT*delT*oscillator[i]*oscillator[i])/(1+delT*damping[i]/2);
		zeta[i]  = (delT*damping[i]/2 - 1) / (delT*damping[i]/2 + 1);
		gamma[i] = (plasmon[i]*plasmon[i]*EPS0*delT*delT) / (delT*damping[i]/2 + 1);
		sumGamma += gamma[i];

		// Set the initial currents to zero
		Jx[i] = 0.0;
		Jy[i] = 0.0;
		Jz[i] = 0.0;
		Jxold[i] = 0.0;
		Jyold[i] = 0.0;
		Jzold[i] = 0.0;
		Jxnew[i] = 0.0;
		Jynew[i] = 0.0;
		Jznew[i] = 0.0;
	};
	EC1 = 0.5*sumGamma / (2*EPS0 + 0.5*sumGamma) ;
	EC2 = 1-EC1;
	EC3 = delT / EPS0 * (1-EC1);
	HC  = delT / MU0;
};


void LorentzCell::updateEFields(){
	
	double dumX=0.0;
	double dumY=0.0;
	double dumZ=0.0;
	
	for(int i=0;i<numPoles;i++){
		dumX += (alpha[i]+1)*Jx[i] +zeta[i]*Jxold[i];
		dumY += (alpha[i]+1)*Jy[i] +zeta[i]*Jyold[i];
		dumZ += (alpha[i]+1)*Jz[i] +zeta[i]*Jzold[i];
	};
	dumX = 0.5*dumX;
	dumY = 0.5*dumY;
	dumZ = 0.5*dumZ;

	Exnew = EC1*Exold + EC2*Ex + EC3 * ( (Hz-cellBackY->Hz)/delY - (Hy-cellBackZ->Hy)/delZ - dumX - outJx);
	Eynew = EC1*Eyold + EC2*Ey + EC3 * ( (Hx-cellBackZ->Hx)/delZ - (Hz-cellBackX->Hz)/delX - dumY - outJy);
	Eznew = EC1*Ezold + EC2*Ez + EC3 * ( (Hy-cellBackX->Hy)/delX - (Hx-cellBackY->Hx)/delY - dumZ - outJz);

	for(int i=0;i<numPoles;i++){
		Jxnew[i] = alpha[i]*Jx[i] + zeta[i]*Jxold[i] + gamma[i]*(Exnew - Exold)/(2*delT);
		Jynew[i] = alpha[i]*Jy[i] + zeta[i]*Jyold[i] + gamma[i]*(Eynew - Eyold)/(2*delT);
		Jznew[i] = alpha[i]*Jz[i] + zeta[i]*Jzold[i] + gamma[i]*(Eznew - Ezold)/(2*delT);
	}

	Exold = Ex;
	Eyold = Ey;
	Ezold = Ez;
	Ex = Exnew;
	Ey = Eynew;
	Ez = Eznew;

	for(int i=0;i<numPoles;i++){
		Jxold[i]=Jx[i];
		Jyold[i]=Jy[i];
		Jzold[i]=Jz[i];
		Jx[i] = Jxnew[i];
		Jy[i] = Jynew[i];
		Jz[i] = Jznew[i];
	}


};


// Usual cell update equations
void LorentzCell::updateHFields(){
	Hx = Hx + HC * ( (cellFrontZ->Ey-Ey)/delZ - (cellFrontY->Ez-Ez)/delY );
	Hy = Hy + HC * ( (cellFrontX->Ez-Ez)/delX - (cellFrontZ->Ex-Ex)/delZ );
	Hz = Hz + HC * ( (cellFrontY->Ex-Ex)/delY - (cellFrontX->Ey-Ey)/delX );
};

// Note that the following functions (addJx, addJy, addJz) should be called before updateEFields() function
void LorentzCell::addJx(double outJx){
	this->outJx = outJx;	
};

void LorentzCell::addJy(double outJy){
	this->outJy = outJy;	
};

void LorentzCell::addJz(double outJz){
	this->outJz = outJz;	
};

void LorentzCell::addMx(double Mx){
	Hx = Hx - HC * Mx;
};

void LorentzCell::addMy(double My){
	Hy = Hy - HC * My;
};

void LorentzCell::addMz(double Mz){
	Hz = Hz - HC * Mz;
};
