#ifndef LorentzCellGUARD
#define LorentzCellGUARD

#include "Cell.h"
// #include "Sourcable.h"

// This cell implements the Lorentz type multipole dispersive materials in the FDTD domain. For details refer to:

/*
Title:	Simple treatment of multi-term dispersion in FDTD
Author:	Okoniewski, M ; Mrozowski, M ; Stuchly, MA
Journal:	IEEE MICROWAVE AND GUIDED WAVE LETTERS; MAY 1997; v.7, no.5, p.121-123
*/

// And also refer to pp 394-397 at:
/*
				  ISBN: 1580530761
                Author: Taflove, Allen.
                 Title: Computational electrodynamics : the finite-difference
                        time-domain method.
               Edition: 2nd ed. / Allen Taflove, Susan C. Hagness.
               Imprint: Boston : Artech House, c2000.

*/

// The code implements the Lorentz Auxiliary Differential Equations - Synchronized (LADES) method as described in the references above.
// LADES requires back storage, but is more stable.


class LorentzCell: public Cell{
public:
	// Constructor method
	
	LorentzCell(int numPoles, const double *plasmon, const double *oscillator, const double *damping);

	// Field update equations for the class
	void updateEFields();
	void updateHFields();

	// Electric and Magnetic Current Source Functions
	void addJx(double Jx);
	void addJy(double Jy);
	void addJz(double Jz);
	void addMx(double Mx);
	void addMy(double My);
	void addMz(double Mz);
private:
	int numPoles;					// Number of poles in the Lorentz model
	double Exold, Eyold, Ezold;		// Previous time step's E field values
	double Exnew, Eynew, Eznew;		// Next time step's E field values
	double outJx, outJy, outJz;		// External current source term
	double *Jx, *Jy, *Jz;			// Currents for each pole
	double *Jxold, *Jyold, *Jzold;	// Previous time step's current values	
	double *Jxnew, *Jynew, *Jznew;	// Next time step's current values	
	double *alpha, *zeta, *gamma;	// Parameters of the Lorentz material
	double EC1, EC2, EC3;			// Constants to be used in E field update equations
	double HC;						// Constant to be used in H field update equations

};

#endif