% This file is responsible for creating the numerical simulation domain
% which will be crunched by the C++ program. It generates the MAT
% file that has the parameters of simulation and the properties of the
% simulation medium.

% This program generates an empty world of free space
% surrounded by PML media. It also defines an excitation which is to be
% sourced at the center of the numerical domain.

clear all
close all

% Global variables
global objectMapIn
global objectColorIn

% Name of the input file
INPUTFILENAME = './SimFiles/SleeveDipole.in';
% Name of the output mat file
OUTPUTFILENAME = './SimFiles/SleeveDipole.mat';
% Name of the mat file that has the info for post simulation calculations
POSTSIMFILENAME = './SimFiles/postSim.mat';

% Include functions in subdirectories
addpath(genpath(pwd));

% Load various constants in MKS units
physicalConstants

% Parse the input file to read the necessary simulation parameters

in=parseInput(INPUTFILENAME);

unitLength = in.unitLength;
delX = in.dx*unitLength;
delY = in.dy*unitLength;
delZ = in.dz*unitLength;
sizeX = in.sizeX*unitLength;
sizeY = in.sizeY*unitLength;
sizeZ = in.sizeZ*unitLength;

delT= 0.9 / ( PC.c*sqrt(delX^-2 + delY^-2 + delZ^-2)); % 0.9 times the Courant stability limit

% Number of cells around the origin is found by calculating the number of
% cells sizeX/2 cuts through. Even if there's a tiny intersection with the
% boundary, as long as there is an intersection (of 1%) , the cell is chosen to
% exist.
NX = 2*ceil(0.5*sizeX / delX - 0.01);
NY = 2*ceil(0.5*sizeY / delY - 0.01);
NZ = 2*ceil(0.5*sizeZ / delZ - 0.01);

% Create the objectMapIn
% objectMapIn is first filled fully with the firstly defined material
objectMapIn = cell(NX,NY,NZ);
objectMapIn(1:NX,1:NY,1:NZ)={in.material{1}};

% objectColorIn holds the reference to the material properties
% The first material fills the entire simulation domain and has the
% index=1.
objectColorIn = ones(NX,NY,NZ);

% depending on the objects, modify the objectMapIn by calling createObjects

% Create the layers

% First create the layers in 3D space,get a slice, add PML's to the top and bottom and save in
% objectMapLayer2D
numPMLLayer = 16; % number of PML cells at the z boundary
numStretchLayer = 10; % How many cells to add around the original input

createObjects(in,'layer');
objectMap2DLayer = objectMapIn(:,NY/2,:);
objectColor2DLayer = objectColorIn(:,NY/2,:);
% Creation of the output points
if length(in.outputLayer)~=0
    objectMap2DLayer_outputIndices = createOutputIndices(in.outputLayer,in,objectColorIn);
end
% Expansion of the simulation domain and addition of PMLs around it
[objectMap2DLayer,objectColor2DLayer] = stretchXZ(numStretchLayer,numStretchLayer,objectMap2DLayer,objectColor2DLayer);
[objectMap2DLayer,objectColor2DLayer,dum]=addPML([0 0 numPMLLayer],objectMap2DLayer,objectColor2DLayer,in);
clear dum
% Modification of the output indices that take into account the expansion
% and the PMLs
if exist('objectMap2DLayer_outputIndices')
    objectMap2DLayer_outputIndices(1,:) = objectMap2DLayer_outputIndices(1,:) + numStretchLayer; % x indices, note that no PMLs in x dir
    objectMap2DLayer_outputIndices(3,:) = objectMap2DLayer_outputIndices(3,:) + numStretchLayer + numPMLLayer; % z indices 
    objectMap2DLayer_outputIndices(2,:) = 1; % the y coordinates should all be 1
end

% Evaluate the line below to visualize objectMap2DLayer
if exist('objectMap2DLayer_outputIndices')
    visualizeLayer(objectColor2DLayer,objectMap2DLayer_outputIndices);
else
    visualizeLayer(objectColor2DLayer);
end

% Put the wire in and do the same as above. This time put PMLs at x and z
numPMLWire = 10; % Number of PML layers around the x and z boundaries
numStretchWire = 6;% Number of cells around the original input 

createObjects(in,'wire');
objectMap2DWire = objectMapIn(:,NY/2,:);
objectColor2DWire = objectColorIn(:,NY/2,:);

% Creation of the output points
if length(in.outputWire)~=0
    objectMap2DWire_outputIndices = createOutputIndices(in.outputWire,in,objectColorIn);
end
% Expansion of the simulation domain and addition of PMLs around it
[objectMap2DWire,objectColor2DWire] = stretchXZ(numStretchWire,numStretchWire,objectMap2DWire,objectColor2DWire);
[objectMap2DWire,objectColor2DWire,dum]=addPML([numPMLWire 0 numPMLWire],objectMap2DWire,objectColor2DWire,in);
clear dum
% Modification of the output indices that take into account the expansion
% and the PMLs
if exist('objectMap2DWire_outputIndices')
    objectMap2DWire_outputIndices(1,:) = objectMap2DWire_outputIndices(1,:) + numStretchWire + numPMLWire; % x indices, note that this time PMLs exist around x
    objectMap2DWire_outputIndices(3,:) = objectMap2DWire_outputIndices(3,:) + numStretchWire + numPMLWire; % z indices 
    objectMap2DWire_outputIndices(2,:) = 1; % The y coordinates should all be 1
end
% Evaluate the line below to visualize objectMap2DWire
if exist('objectMap2DWire_outputIndices')
    visualizeLayer(objectColor2DWire,objectMap2DWire_outputIndices);
else
    visualizeLayer(objectColor2DWire);
end

% And finally add the final antenna
numPMLAntenna = 10; % Number of PML layers in 3D simulation, no expansion is done. 

createObjects(in,'antenna');

% Creation of the output indices
if length(in.outputAntenna)~=0
    objectMap3D_outputIndices = createOutputIndices(in.outputAntenna,in,objectColorIn);
end
% Addition of the PML layers
[objectMap3D,objectColor3D,material]=addPML([numPMLAntenna numPMLAntenna numPMLAntenna],objectMapIn,objectColorIn,in);
in.material = material;
% Modification of the output indices to take into account the addition of
% PMLS
if exist('objectMap3D_outputIndices')
    objectMap3D_outputIndices(:,:) = objectMap3D_outputIndices(:,:) + numPMLAntenna; % PMLs around all boundaries
end

% Evaluate the line below to visualize objectMap3D
if exist('objectMap3D_outputIndices')
    visualizeWorld(in.material,objectColor3D,objectMap3D_outputIndices);
else
    visualizeWorld(in.material,objectColor3D);
end


% If TE > 0 simulation is in TE, if TE < 0 simulation is TM
if (in.isTE > 0)
    TE = 10;
else
    TE = -10;
end

% Creation of the source
source = getSource(in.lambda*in.unitLength,delT,in.delF,in.numCycles);


disp('Saving output file... It''ll take a while.');
execute=sprintf('save ''%s'' delX delY delZ delT TE source objectMap3D objectMap2DLayer objectMap2DWire',OUTPUTFILENAME);

% If the ouput indices exist, add them to the output file
if exist('objectMap2DLayer_outputIndices')
    execute = [execute ' objectMap2DLayer_outputIndices'];
end

if exist('objectMap2DWire_outputIndices')
    execute = [execute ' objectMap2DWire_outputIndices'];
end

if exist('objectMap3D_outputIndices')
    execute = [execute ' objectMap3D_outputIndices'];
end

% Save without compression
execute = [execute ' -v6'];

% Save output
evalin('base',execute);

% Save the post simulation data file
% Add all the variables needed for post processing below
disp('Saving post simulation file...')
save(POSTSIMFILENAME,'objectMap2DLayer_outputIndices','objectMap2DWire_outputIndices','objectMap3D_outputIndices','delT','in','source');
return



