// OOFDTD.cpp : Defines the entry point for the console application.

// This code implements the Finite Difference Time Domain (FDTD) technique using the 
// object oriented features of C++. The aim is to have a code which can easily be extended
// to include different cell types.

#include "mainHeader.h"


int main(int argc, char* argv[]){

	World2D *layer;		// 2D world with only layers
	World2D *wire;		// 2D world with a wire embedded in layers
	World3D *final;		// 3D world with the final structure
	
	// 3D final world sizes without PMLs
	int sizeFinal_x,sizeFinal_y,sizeFinal_z;
	
	// 2D layer world sizes without PMLs
	int sizeLayer_x, sizeLayer_z;

	// 2D wire world sizes without PMLs
	int sizeWire_x, sizeWire_z;

	if (argc != 2){
		printf("Usage: oofdtd <matfile>");
		printf(" where <matfile> is the name of the MAT-file");
		printf(" that has the simulation parameters.\n");
		return 0;
	}
	
	// Okay here we go. First set up the 2D Source World, Layer
	layer = new World2D(argv[1],"objectMap2DLayer","layer.out");
	layer->addPeriodicBoundaryBricks();
	layer->sanityCheck();
	
	// Now set up the wire world
	wire = new World2D(argv[1],"objectMap2DWire","wire.out");
	wire->addBoundaryBricks();
	wire->sanityCheck();
	
	// Time for the 3d World, Final
	final = new World3D(argv[1],"objectMap3D","antenna.out");
	final->addBoundaryBricks();
	final->sanityCheck();

	
	// Coordinate transformation variables
	// Size of the 2D Layer
	sizeLayer_x = layer->NX - 2*layer->numPMLx;
	sizeLayer_z = layer->NZ - 2*layer->numPMLz;

	// Size of the 2D Wire
	sizeWire_x = wire->NX - 2*wire->numPMLx;
	sizeWire_z = wire->NZ - 2*wire->numPMLz;
	
	// Size of the 3D box which holds the simulation domain:
	sizeFinal_x = final->NX - 2*final->numPMLx;
	sizeFinal_y = final->NY - 2*final->numPMLy;
	sizeFinal_z = final->NZ - 2*final->numPMLz;

	// Sanity check
	if((sizeWire_x - sizeFinal_x)%2!=0 || (sizeLayer_x - sizeWire_x)%2!=0 ){
		printf("Broken symmetry. Size difference is not zero modulo 2. Exiting...");
		exit(EXIT_FAILURE);
	}
	
	// Location of the bottom and top corner points of plain 3D world in the wire world
	int botWire_x =wire->numPMLx + (sizeWire_x - sizeFinal_x)/2 ; 
	int botWire_z =wire->numPMLz + (sizeWire_z - sizeFinal_z)/2 ;
	int topWire_x =botWire_x + sizeFinal_x - 1; 
	int topWire_z =botWire_z + sizeFinal_z - 1; 
	
	// Location of the bottom and top corner points of plain 3D world in the layer world
	int botLayer_x =layer->numPMLx + (sizeLayer_x - sizeFinal_x)/2 ; 
	int botLayer_z =layer->numPMLz + (sizeLayer_z - sizeFinal_z)/2 ;
	int topLayer_x =botLayer_x + sizeFinal_x - 1; 
	int topLayer_z =botLayer_z + sizeFinal_z - 1; 


	// Location of the bottom and top corner points of plain 3D world in the 3D world with PMLs
	int botFinal_x = final->numPMLx;
	int topFinal_x = botFinal_x + sizeFinal_x -1;
	int botFinal_y = final->numPMLy;
	int topFinal_y = botFinal_y + sizeFinal_y -1;
	int botFinal_z = final->numPMLz;
	int topFinal_z = botFinal_z + sizeFinal_z -1;


	// Shifts in indices going from wire to final
	int delWireToFinal_x = botFinal_x - botWire_x;
	int delWireToFinal_z = botFinal_z - botWire_z;

	// Shifts in indices going from layer to wire
	int delLayerToWire_x = botWire_x - botLayer_x;
	int delLayerToWire_z = botWire_z - botLayer_z;
	
	// Distance from the plain 3D boundaries for the two transportation windows
	int outWinDist = 2;
	int inWinDist = 1;

	
	// Set up the sources for each world		
	wire->setSourceWorld(layer,botLayer_x-outWinDist,topLayer_x+outWinDist+1,botLayer_z-outWinDist,topLayer_z+outWinDist+1,delLayerToWire_x,delLayerToWire_z);
	final->setSourceWorld(wire,botWire_x+inWinDist,topWire_x-inWinDist+1,botFinal_y+inWinDist,topFinal_y-inWinDist+1,botWire_z+inWinDist,topWire_z-inWinDist+1,delWireToFinal_x,0,delWireToFinal_z);

#ifndef NOVISUAL
	// Initialization of the visualization routines
	if(layer->isTE){
		final->initVisualizer("Ey","y",61-1,"finalField"); // Determine the slice number from Matlab visualization. Subtract one to take into account the 0 based indexing of C++.
		wire->initVisualizer("Ey","y",(wire->NY -1),"wireField");
		layer->initVisualizer("Ey","y",(layer->NY -1),"layerField");
	}
	else{
		final->initVisualizer("Hy","y",61-1,"finalField");
		wire->initVisualizer("Hy","y",(wire->NY -1),"wireField");
		layer->initVisualizer("Hy","y",(layer->NY -1),"layerField");
	}

#endif

	
	time_t start,now;
	time(&start);
	double dif;
	
	printf("Starting the time loop...\n");
	while(layer->currentTime < layer->totalTimeSteps){			

		layer->propagateE();
		if(layer->isTE)
			layer->TFSourceE();
		
		// Note that for dispersive media sourcing should be done before updating the 
		// fields. Thus TFSFSourceE and TFSFSourceH functions should be called before the
		// propagateE or propagateH functions.
		wire->TFSFsourceH();
		wire->propagateH();
		
		final->TFSFsourceE(wire->isTE);
		final->propagateE();		

		layer->propagateH();
		
		if(!layer->isTE)
            layer->TFSourceH();

		wire->TFSFsourceE();
		wire->propagateE();		
		
		final->TFSFsourceH(wire->isTE);
		final->propagateH();		

		layer->currentTime++;
		wire->currentTime++;
		final->currentTime++;

		layer->outputFields();
		wire->outputFields();
		final->outputFields();

		if(layer->currentTime%1000==0){
			time(&now);
			dif = difftime(now,start); // Elapsed time in seconds
			dif = dif  /60; // Conversion to minutes
			printf("Time Step: %d/%d ** Elapsed Time: %.2lf mins ** Remaining Time: %.2lf mins ** Total time: %.2lf mins\n",layer->currentTime,layer->totalTimeSteps,dif,dif/layer->currentTime*layer->totalTimeSteps - dif,dif/layer->currentTime*layer->totalTimeSteps);
		}

#ifndef NOVISUAL
		if(layer->currentTime%10==0){
			layer->updateVisualization();
			wire->updateVisualization();
			final->updateVisualization();
		}
#endif

	}
};
