#include "World3D.h"

// Static variable initialization

#ifndef NOVISUAL
Engine* World3D::ep=0;
#endif


// Constructor
World3D::World3D(const char *inputFile, const char *varName,const char *outputFile){
	NX=NY=NZ=0;
	numPMLx=numPMLy=numPMLz=0;
	delX=delY=delZ=delT=0.0;
	totalTimeSteps=currentTime=0;
	objectMap3D=0;
	brickMap=0;
	numBricks=0;
	
	// TFSF field teleportation limit anf offset variables
	xminTFSF=xmaxTFSF=yminTFSF=ymaxTFSF=zminTFSF=zmaxTFSF=x0TFSF=y0TFSF=z0TFSF=0;

	// The lists that hold the indices of cells to be outputted
	fp = 0; // output file pointer
	numOutputCells=0;
	ilist=0;
	jlist=0;
	klist=0;
	
	// Initialize all the parameters
	readMAT(inputFile,varName);
	// Set the 6 neighbours of each of the cells
	setNeighbours();
	
	// See how many PMLs you have in x,y,z directions
	setnumPML();

	if(outputFile != 0){
		fp = fopen(outputFile,"wb+");

		// The indices for a specific world always should have the following format
		char buf[200];
		strcpy(buf,varName);
		strcat(buf,"_outputIndices");
		getOutputIndices(inputFile, buf);
	}
}


void World3D::getOutputIndices(const char *file, const char *varName){
	MATFile *pmat;	// pointer to the MAT file
	mxArray *pa;	// pointer to array that holds indices
	double *pr;		// pointer to real part of pa

	printf("Reading file %s...\n\n", file);

	// Open file
	pmat = matOpen(file, "r");
	if (pmat == NULL) {
		printf("Error opening file %s\n", file);
		exit(EXIT_FAILURE);
	}

	pa = matGetVariable(pmat,varName);
	if(pa==0){
		printf("Variable %s not found in %s. No outputting will be done.\n",varName,file);
		fclose(fp);
		fp=0;
		return;
	}
	int noDim = mxGetNumberOfDimensions(pa);
	const int *dims = mxGetDimensions(pa);

	if(noDim != 2){
		printf("Variable %s is not 2D. Exiting...\n",varName);
		exit(EXIT_FAILURE);
	}

	if(dims[0]!=3){
		printf("Variable %s does not have 3 rows. Exiting...\n",varName);
		exit(EXIT_FAILURE);
	}
	
	numOutputCells = dims[1];
	printf("There are %d output cells in %s.\n",numOutputCells,varName);

	ilist = new int[numOutputCells];
	jlist = new int[numOutputCells];
	klist = new int[numOutputCells];

	pr = mxGetPr(pa);
	// Remember columns are saved first in Matlab
	for(int m=0;m<numOutputCells;m++){
		ilist[m]=(int)pr[3*m];
		jlist[m]=(int)pr[3*m +1];
		klist[m]=(int)pr[3*m +2];
	}


	// Close the file
	if (matClose(pmat) != 0) {
		printf("Error closing file %s\n",file);
		exit(EXIT_FAILURE);
	}


}
// This function gets the scalar "name" (either int or double) from the MAT file and puts it
// into the memory location pointed by num
template<class T> void World3D::getScalar(MATFile *pmat, const char* name, T *num){
	mxArray *pa;
	pa = matGetVariable(pmat,name);
	if (pa==0){
		printf("Variable %s not found in the file. Exiting...\n",name);
		exit(EXIT_FAILURE);
	}
	(*num) = (T) mxGetScalar(pa);
	printf("%s = %g \n",name,(double) *num);
	mxDestroyArray(pa);
}


// Reads the object definitions from MAT file and instantiates the objects
// As new object types are added, this function needs to be modified

void World3D::getObjectMap3D(MATFile *pmat,const char *varName){
	mxArray *pa;	// array pointer
	mxArray *ps;	// structure pointer
	char *type=new char[100]; // string buffer that holds the name of the type of cell
	int noDim;			// Number of dimentions
	const int *dims;	// Number of elements in each dimension
	int totalSize;		// total number of cells

	// Okay, now read the objectMap3D
	pa = matGetVariable(pmat,varName);
	if(pa==0){
		printf("Variable %s not found in the file. Exiting...\n",varName);
		exit(EXIT_FAILURE);
	}
	noDim = mxGetNumberOfDimensions(pa);
	dims = mxGetDimensions(pa);

	// Check dimensions
	if(noDim!=3){
		printf("Number of dimensions for %s is %d. It should have been 3. Exiting...",varName,noDim);
		exit(EXIT_FAILURE);
	};

	// Get the dimensions of objectMap3D
	printf("%s is of size ", varName);
	totalSize = 1;
	for(int i=0;i<noDim;i++){ 
		printf("%d ",dims[i]);
		totalSize = totalSize * dims[i];
	}
	printf(" = %d\n",totalSize);
	NX = dims[0];
	NY = dims[1];
	NZ = dims[2];


	// Begin filling the objectMap3D
	objectMap3D = new Cell*[NX*NY*NZ];

	for (int i=0;i<totalSize;i++){
		ps=mxGetCell(pa,i); // Note that Matlab saves data in column major numbering system
		if (mxIsStruct(ps)){
			// Get the type of cell
			mxGetString(mxGetField(ps,0,"type"),type,50);
			if( strcmp(type,"DielectricECell")== 0){
				double epsr = mxGetScalar(mxGetField(ps,0,"epsr"));
				objectMap3D[i]=new DielectricECell(epsr);
			}else if(strcmp(type,"PMLCell")== 0){
				double Sx= mxGetScalar(mxGetField(ps,0,"Sx"));
				double Sy= mxGetScalar(mxGetField(ps,0,"Sy"));
				double Sz= mxGetScalar(mxGetField(ps,0,"Sz"));
				double Kx= mxGetScalar(mxGetField(ps,0,"Kx"));
				double Ky= mxGetScalar(mxGetField(ps,0,"Ky"));
				double Kz= mxGetScalar(mxGetField(ps,0,"Kz"));
				double epsr = mxGetScalar(mxGetField(ps,0,"epsr"));
				objectMap3D[i]=new PMLCell(Sx,Sy,Sz,Kx,Ky,Kz,epsr);
			}else if(strcmp(type,"LorentzCell")== 0){
				int numPoles = (int) mxGetScalar(mxGetField(ps,0,"numPoles"));
				const double *plasmon = mxGetPr(mxGetField(ps,0,"plasmon"));
				const double *oscillator  = mxGetPr(mxGetField(ps,0,"oscillator"));
				const double *damping = mxGetPr(mxGetField(ps,0,"damping"));
				objectMap3D[i]=new LorentzCell(numPoles,plasmon,oscillator,damping);
			}else {
				printf("%s not implemented. Exiting...\n",type);
				exit(EXIT_FAILURE);
			}			
		}else{
			printf("Cell element %d is not a struct. Something's wrong. Exiting...\n",i);
			exit(EXIT_FAILURE);
		}
	};
	mxDestroyArray(pa);

}


void World3D::getSource(MATFile *pmat,const char *sourceName){
	mxArray *pa;	// array pointer
	int noDim;			// Number of dimentions
	const int *dims;	// Number of elements in each dimension
	double *paReal;	// real point of pa

	pa = matGetVariable(pmat,sourceName);
	if(pa==0){
		printf("Variable source not found in the file. Exiting...\n");
		exit(EXIT_FAILURE);
	}
	noDim = mxGetNumberOfDimensions(pa);
	dims = mxGetDimensions(pa);
	if (noDim !=2 || (dims[0]==1 && dims[1]==1) || (dims[0]>1 && dims[1]>1) ){
		printf("Dimension mismatch for source. Exiting... \n");
		exit(EXIT_FAILURE);
	}
	printf("Source is of length %d \n",dims[0]*dims[1]);
	totalTimeSteps = dims[0]*dims[1];

	fieldSource = new double[dims[0]*dims[1]];
	paReal = mxGetPr(pa);
	// get each element one by one....
	for (int i=0;i<dims[0]*dims[1];i++){
		fieldSource[i]=paReal[i];
	};
	mxDestroyArray(pa);
}
// Opens the MAT file and reads in the simulations parameters
void World3D::readMAT(const char *file, const char *varName) {

	MATFile *pmat;	// pointer to the MAT file

	printf("Reading file %s...\n\n", file);

	// Open file
	pmat = matOpen(file, "r");
	if (pmat == NULL) {
		printf("Error opening file %s\n", file);
		exit(EXIT_FAILURE);
	}

	// Read delX, delY, delZ, delT
	getScalar(pmat,"delX",&delX);
	getScalar(pmat,"delY",&delY);
	getScalar(pmat,"delZ",&delZ);
	getScalar(pmat,"delT",&delT);

	// Set the simulation parameters DX,DY,DZ,DT
	// Note that these parameters should be set before any Cell objects are instantiated
	// Otherwise, the coefficient updates made in the construction of objects will be invalid
	Cell::delX=delX;
	Cell::delY=delY;
	Cell::delZ=delZ;
	Cell::delT=delT;

	// Read the source array
	getSource(pmat,"source");


	// Read the cells
	getObjectMap3D(pmat,varName);

	// Close the file
	if (matClose(pmat) != 0) {
		printf("Error closing file %s\n",file);
		exit(EXIT_FAILURE);
	}

	printf("Done.\n");	
}



void World3D::addBoundaryBricks(){

	printf("Adding boundary bricks for World3D.\n");

	const int imin=0,imax=NX-1,jmin=0,jmax=NY-1,kmin=0,kmax=NZ-1;
	numBricks=2*(NX*NY+NX*NZ+NY*NZ);
	brickMap = new Cell*[numBricks];
	int ib = 0;


	// Top face z=max(z)
	for(int i=0;i<NX;i++){
		for(int j=0;j<NY;j++){
			ObjectMap3D(i,j,kmax)->cellFrontZ=new PECBrick();
			brickMap[ib]=ObjectMap3D(i,j,kmax)->cellFrontZ;
			brickMap[ib]->cellBackZ=ObjectMap3D(i,j,kmax);
			ib++;
		}
	}

	// Bottom face z=min(z)
	for(int i=0;i<NX;i++){
		for(int j=0;j<NY;j++){
			ObjectMap3D(i,j,kmin)->cellBackZ=new PECBrick();
			brickMap[ib]=ObjectMap3D(i,j,kmin)->cellBackZ;
			brickMap[ib]->cellFrontZ=ObjectMap3D(i,j,kmin);
			ib++;
		}
	}

	// Left face x=min(x)
	for(int k=0;k<NZ;k++){
		for(int j=0;j<NY;j++){
			ObjectMap3D(imin,j,k)->cellBackX=new PECBrick();
			brickMap[ib]=ObjectMap3D(imin,j,k)->cellBackX;
			brickMap[ib]->cellFrontX=ObjectMap3D(imin,j,k);
			ib++;
		}
	}

	// Right face x=max(x)
	for(int k=0;k<NZ;k++){
		for(int j=0;j<NY;j++){
			ObjectMap3D(imax,j,k)->cellFrontX=new PECBrick();
			brickMap[ib]=ObjectMap3D(imax,j,k)->cellFrontX;
			brickMap[ib]->cellBackX=ObjectMap3D(imax,j,k);
			ib++;
		}
	}

	// Front face y=min(y)
	for(int k=0;k<NZ;k++){
		for(int i=0;i<NX;i++){
			ObjectMap3D(i,jmin,k)->cellBackY=new PECBrick();
			brickMap[ib]=ObjectMap3D(i,jmin,k)->cellBackY;
			brickMap[ib]->cellFrontY=ObjectMap3D(i,jmin,k);
			ib++;
		}
	}

	// Back face y=max(y)
	for(int k=0;k<NZ;k++){
		for(int i=0;i<NX;i++){
			ObjectMap3D(i,jmax,k)->cellFrontY=new PECBrick();
			brickMap[ib]=ObjectMap3D(i,jmax,k)->cellFrontY;
			brickMap[ib]->cellBackY=ObjectMap3D(i,jmax,k);
			ib++;
		}
	}

}



// Set the neighbours of each cell
void World3D::setNeighbours(){

	// Update cellFrontX and cellBackX	
	for(int i=1;i<NX;i++){
		for(int j=0;j<NY;j++){
			for(int k=0;k<NZ;k++){
				ObjectMap3D(i,j,k)->cellBackX = ObjectMap3D(i-1,j,k);
				ObjectMap3D(i-1,j,k)->cellFrontX = ObjectMap3D(i,j,k);
			}
		}
	}

	// Update cellFrontY and cellBackY	
	for(int i=0;i<NX;i++){
		for(int j=1;j<NY;j++){
			for(int k=0;k<NZ;k++){
				ObjectMap3D(i,j,k)->cellBackY = ObjectMap3D(i,j-1,k);
				ObjectMap3D(i,j-1,k)->cellFrontY = ObjectMap3D(i,j,k);
			}
		}
	}

	// Update cellFrontZ and cellBackZ	
	for(int i=0;i<NX;i++){
		for(int j=0;j<NY;j++){
			for(int k=1;k<NZ;k++){
				ObjectMap3D(i,j,k)->cellBackZ = ObjectMap3D(i,j,k-1);
				ObjectMap3D(i,j,k-1)->cellFrontZ = ObjectMap3D(i,j,k);
			}
		}
	}

}





// Set the numPML variables
void World3D::setnumPML(){
	PMLCell *pml;
	// set numPMLx
	numPMLx = 0;
	for(int i=0;i<NX;i++){
		pml = dynamic_cast<PMLCell*>(ObjectMap3D(i,NY/2,NZ/2));
		if(pml!=0)
			numPMLx++;
		else
			break;
	}

	// set numPMLy
	numPMLy = 0;
	for(int j=0;j<NY;j++){
		pml = dynamic_cast<PMLCell*>(ObjectMap3D(NX/2,j,NZ/2));
		if(pml!=0)
			numPMLy++;
		else
			break;
	}
	
	// set numPMLz
	numPMLz = 0;
	for(int k=0;k<NZ;k++){
		pml = dynamic_cast<PMLCell*>(ObjectMap3D(NX/2,NY/2,k));
		if(pml!=0)
			numPMLz++;
		else
			break;
	}

	printf("Number of PMLs in x-y-z: %d %d %d\n",numPMLx,numPMLy,numPMLz);


}

// Sanity check before computation starts
void World3D::sanityCheck(){

	for(int i=0;i<NX*NY*NZ;i++){
		// Check whether the fundamental simulation variables are set correctly
		if(objectMap3D[i]->delX != delX || objectMap3D[i]->delY != delY || objectMap3D[i]->delZ != delZ || objectMap3D[i]->delT != delT){
			printf("Objects with wrong parameters exist. Exiting...");
			exit(EXIT_FAILURE);
		}
		// Check whether all neighbours of all cells have been set
		objectMap3D[i]->checkNeighbours((i%(NX*NY))%NX,(i%(NX*NY))/NX,i/(NX*NY));
	}

	// Check brick cells 
	for(int i=0;i<numBricks;i++)
		(brickMap[i])->checkNeighbours(i);


}


void World3D::propagateE(){

	// Call the updateEFields() function of the computational boundary cells
	for(int i=0;i<NX*NY*NZ;i++){
		objectMap3D[i]->updateEFields();
	}

	// Set tangential fields to zero
	for(int i=0;i<numBricks;i++){
		brickMap[i]->updateEFields();
	}
}

void World3D::TFSourceE(){
	// Source the E-Fields 1 layer away from top PML
	for(int i=0;i<NX;i++)
		ObjectMap3D(i,0,NZ-numPMLz-2)->Ey += fieldSource[currentTime];
}

void World3D::TFSourceH(){
	// Source the H-Fields 1 layer away from top PML
	for(int i=0;i<NX;i++)
		ObjectMap3D(i,0,NZ-numPMLz-2)->Hy += fieldSource[currentTime];
}

void World3D::propagateH(){
	// Call the updateHFields() function of the computational boundary cells
	for(int i=0;i<NX*NY*NZ;i++){
		objectMap3D[i]->updateHFields();
	}

	// Set tangential fields to zero
	for(int i=0;i<numBricks;i++){
		brickMap[i]->updateHFields();
	}
}


void World3D::setSourceWorld(World3D *source,int xmin,int xmax,int ymin, int ymax, int zmin,int zmax, int x0, int y0, int z0){
	sourceTF = source;
	xminTFSF = xmin;
	xmaxTFSF = xmax;
	yminTFSF = ymin;
	ymaxTFSF = ymax;
	zminTFSF = zmin;
	zmaxTFSF = zmax;
	x0TFSF = x0;
	y0TFSF = y0;
	z0TFSF = z0;
}


// Sourcing a World3D with a World2D

// Below is a simple script to get the shifted (i,j,k)th term of objectMap
// Because matlab stores multidimensional arrays in column major format, so does objectMap.
// objectMap[k*NX*NY+j*NX+i] gives the cell (i,j,k).
// ShiftedObjectMap3D is used so as to not to modify the arguments one by one
#define ShiftedObjectMap3D(I,J,K) objectMap3D[(K+z0TFSF)*NX*NY+(J+y0TFSF)*NX + (I+x0TFSF)]

void World3D::TFSFsourceE(bool isTE){
	if(sourceTF == 0){
		printf("sourceTF called with no pointer to source. Exiting...");
		exit(EXIT_FAILURE);
	}
	// A very bad programming choice has led to a bug that needs the following fix
	// Note that the macro ObjectMap3D defined in World3D.h has
	// ObjectMap3D(I,J,K) objectMap3D[(K)*NX*NY+(J)*NX + (I)]
	// note that when one does sourceTF->ObjectMap3D(i,0,zminTFSF)
	// the NX in the macro is *this.NX, not *sourceTF.NX . That's a major bug.
	// But here's the fix. In the future, never call ObjectMap outside of the object

	int sNX = sourceTF->NX;
	int sNY = sourceTF->NY;

#define ObjectMap3DSource(I,J,K) objectMap3D[(K)*sNX*sNY+(J)*sNX + (I)]

	if(isTE){
		// z=min(z) boundary - xy face
		for(int j=yminTFSF;j<ymaxTFSF;j++)
			for(int i=xminTFSF+1;i<xmaxTFSF;i++)			
				ShiftedObjectMap3D(i,j,zminTFSF)->addJy(+(sourceTF->ObjectMap3DSource(i,0,zminTFSF)->Hx)/delZ);

		// z=max(z) boundary
		for(int j=yminTFSF;j<ymaxTFSF;j++)
			for(int i=xminTFSF+1;i<xmaxTFSF;i++)			
				ShiftedObjectMap3D(i,j,zmaxTFSF)->addJy(-(sourceTF->ObjectMap3DSource(i,0,zmaxTFSF-1)->Hx)/delZ);

		// x=min(x) boundary
		for(int k=zminTFSF+1;k<zmaxTFSF;k++)
			for(int j=yminTFSF;j<ymaxTFSF;j++)
				ShiftedObjectMap3D(xminTFSF,j,k)->addJy(-(sourceTF->ObjectMap3DSource(xminTFSF,0,k)->Hz)/delX);

		// x=max(x) boundary
		for(int k=zminTFSF+1;k<zmaxTFSF;k++)
			for(int j=yminTFSF;j<ymaxTFSF;j++)
				ShiftedObjectMap3D(xmaxTFSF,j,k)->addJy(+(sourceTF->ObjectMap3DSource(xmaxTFSF-1,0,k)->Hz)/delX);

		// y=min(y) boundary
		for(int k=zminTFSF;k<zmaxTFSF;k++)
			for(int i=xminTFSF+1;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,yminTFSF,k)->addJz(-(sourceTF->ObjectMap3DSource(i,0,k)->Hx)/delY);
		
		for(int k=zminTFSF+1;k<zmaxTFSF;k++)
			for(int i=xminTFSF;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,yminTFSF,k)->addJx(+(sourceTF->ObjectMap3DSource(i,0,k)->Hz)/delY);
		
		// y=max(y) boundary
		for(int k=zminTFSF;k<zmaxTFSF;k++)
			for(int i=xminTFSF+1;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,ymaxTFSF,k)->addJz(+(sourceTF->ObjectMap3DSource(i,0,k)->Hx)/delY);

		for(int k=zminTFSF+1;k<zmaxTFSF;k++)
			for(int i=xminTFSF;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,ymaxTFSF,k)->addJx(-(sourceTF->ObjectMap3DSource(i,0,k)->Hz)/delY);


	}else{
		//TM implementation

		// z=min(z) boundary
		for(int j=yminTFSF+1;j<ymaxTFSF;j++)
			for(int i=xminTFSF;i<xmaxTFSF;i++)			
				ShiftedObjectMap3D(i,j,zminTFSF)->addJx(-(sourceTF->ObjectMap3DSource(i,0,zminTFSF)->Hy)/delZ);

		// z=max(z) boundary
		for(int j=yminTFSF+1;j<ymaxTFSF;j++)
			for(int i=xminTFSF;i<xmaxTFSF;i++)			
				ShiftedObjectMap3D(i,j,zmaxTFSF)->addJx(+(sourceTF->ObjectMap3DSource(i,0,zmaxTFSF-1)->Hy)/delZ);

		// x=min(x) boundary
		for(int k=zminTFSF;k<zmaxTFSF;k++)
			for(int j=yminTFSF+1;j<ymaxTFSF;j++)
				ShiftedObjectMap3D(xminTFSF,j,k)->addJz(+(sourceTF->ObjectMap3DSource(xminTFSF,0,k)->Hy)/delX);

		// x=max(x) boundary
		for(int k=zminTFSF;k<zmaxTFSF;k++)
			for(int j=yminTFSF+1;j<ymaxTFSF;j++)
				ShiftedObjectMap3D(xmaxTFSF,j,k)->addJz(-(sourceTF->ObjectMap3DSource(xmaxTFSF-1,0,k)->Hy)/delX);

		// y=min(y) and y=max(y) boundaries do not have electric currents for TM polarization
	}
#undef ObjectMap3DSource
}

// Sourcing a World3D with a World2D
void World3D::TFSFsourceH(bool isTE){
	if(sourceTF == 0){
		printf("sourceTF called with no pointer to source. Exiting...");
		exit(EXIT_FAILURE);
	}
	// A very bad programming choice has led to a bug that needs the following fix
	// Note that the macro ObjectMap3D defined in World3D.h has
	// ObjectMap3D(I,J,K) objectMap3D[(K)*NX*NY+(J)*NX + (I)]
	// note that when one does sourceTF->ObjectMap3D(i,0,zminTFSF)
	// the NX in the macro is *this.NX, not *sourceTF.NX . That's a major bug.
	// But here's the fix. In the future, never call ObjectMap outside of the object

	int sNX = sourceTF->NX;
	int sNY = sourceTF->NY;

#define ObjectMap3DSource(I,J,K) objectMap3D[(K)*sNX*sNY+(J)*sNX + (I)]

	if(isTE){
		// z=min(z) boundary
		for(int j=yminTFSF;j<ymaxTFSF;j++)
			for(int i=xminTFSF+1;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,j,zminTFSF)->addMx(+(sourceTF->ObjectMap3DSource(i,0,zminTFSF)->Ey)/delZ);

		// z=max(z) boundary
		for(int j=yminTFSF;j<ymaxTFSF;j++)
			for(int i=xminTFSF+1;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,j,zmaxTFSF-1)->addMx(-(sourceTF->ObjectMap3DSource(i,0,zmaxTFSF)->Ey)/delZ);

		// x=min(x) boundary
		for(int k=zminTFSF+1;k<zmaxTFSF;k++)
			for(int j=yminTFSF;j<ymaxTFSF;j++)
				ShiftedObjectMap3D(xminTFSF,j,k)->addMz(-(sourceTF->ObjectMap3DSource(xminTFSF,0,k)->Ey)/delX);

		// x=max(x) boundary
		for(int k=zminTFSF+1;k<zmaxTFSF;k++)
			for(int j=yminTFSF;j<ymaxTFSF;j++)
				ShiftedObjectMap3D(xmaxTFSF-1,j,k)->addMz(+(sourceTF->ObjectMap3DSource(xmaxTFSF,0,k)->Ey)/delX);
		
		// y=min(y) and y=max(y) boundaries do not have magnetic currents for TE polarization


	}else{
		//TM implementation
		// z=min(z) boundary
		for(int j=yminTFSF+1;j<ymaxTFSF;j++)
            for(int i=xminTFSF;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,j,zminTFSF)->addMy(-(sourceTF->ObjectMap3DSource(i,0,zminTFSF)->Ex)/delZ);

		// z=max(z) boundary
		for(int j=yminTFSF+1;j<ymaxTFSF;j++)
			for(int i=xminTFSF;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,j,zmaxTFSF-1)->addMy(+(sourceTF->ObjectMap3DSource(i,0,zmaxTFSF)->Ex)/delZ);

		// x=min(x) boundary
		for(int k=zminTFSF;k<zmaxTFSF;k++)
			for(int j=yminTFSF+1;j<ymaxTFSF;j++)
				ShiftedObjectMap3D(xminTFSF,j,k)->addMy(+(sourceTF->ObjectMap3DSource(xminTFSF,0,k)->Ez)/delX);

		// x=max(x) boundary
		for(int k=zminTFSF;k<zmaxTFSF;k++)
			for(int j=yminTFSF+1;j<ymaxTFSF;j++)
				ShiftedObjectMap3D(xmaxTFSF-1,j,k)->addMy(-(sourceTF->ObjectMap3DSource(xmaxTFSF,0,k)->Ez)/delX);

		// y=min(y) boundary
		for(int k=zminTFSF+1;k<zmaxTFSF;k++)
			for(int i=xminTFSF;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,yminTFSF,k)->addMz(+(sourceTF->ObjectMap3DSource(i,0,k)->Ex)/delY);

		for(int k=zminTFSF;k<zmaxTFSF;k++)
			for(int i=xminTFSF+1;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,yminTFSF,k)->addMx(-(sourceTF->ObjectMap3DSource(i,0,k)->Ez)/delY);

		// y=max(y) boundary
		for(int k=zminTFSF+1;k<zmaxTFSF;k++)
			for(int i=xminTFSF;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,ymaxTFSF-1,k)->addMz(-(sourceTF->ObjectMap3DSource(i,0,k)->Ex)/delY);

		for(int k=zminTFSF;k<zmaxTFSF;k++)
			for(int i=xminTFSF+1;i<xmaxTFSF;i++)
				ShiftedObjectMap3D(i,ymaxTFSF-1,k)->addMx(+(sourceTF->ObjectMap3DSource(i,0,k)->Ez)/delY);


	}
// Clears definition of bug fixing macro
#undef ObjectMap3DSource
}


// Clear the definition of ShiftedObjectMap3D now.
#undef ShiftedObjectMap3D


void World3D::outputFields(){
	double dum[6];
	int i,j,k;
	for(int m=0;m<numOutputCells;m++){
		i=ilist[m];j=jlist[m];k=klist[m];
		dum[0]=ObjectMap3D(i,j,k)->Ex;
		dum[1]=ObjectMap3D(i,j,k)->Ey;
		dum[2]=ObjectMap3D(i,j,k)->Ez;
		dum[3]=ObjectMap3D(i,j,k)->Hx;
		dum[4]=ObjectMap3D(i,j,k)->Hy;
		dum[5]=ObjectMap3D(i,j,k)->Hz;
		fwrite(dum,sizeof(double),6,fp);
	}

}

#ifndef NOVISUAL
// Field visualization routines for Matlab
// Note that Matlab uses column-major oredering for arrays
// Thus, dmat[(J)*dims[0] + (I)] gives the (I,J) element of the array
// However the way the loops are written are such that the fields are read in row-major
// order. Therefore, a conversion needs to be done. dum/dims[1]+(dum%dims[1])*dims[0] does that.
void World3D::copyDataToVisualizer(const char *field,int limx1, int limx2,int limy1,int limy2, int limz1,int limz2){
	int dum=0;
	
	// dMat[m][n] gives dims[1]*m + n element
	if(strcmp(field,"Ex")==0){
		for(int i=limx1;i<limx2;i++){
			for(int j=limy1;j<limy2;j++){
				for(int k=limz1;k<limz2;k++){
					dMat[dum/dims[1]+(dum%dims[1])*dims[0]]=ObjectMap3D(i,j,k)->Ex;
					dum++;
				}
			}
		}
	}else if(strcmp(field,"Ey")==0){
		for(int i=limx1;i<limx2;i++){
			for(int j=limy1;j<limy2;j++){
				for(int k=limz1;k<limz2;k++){
					dMat[dum/dims[1]+(dum%dims[1])*dims[0]]=ObjectMap3D(i,j,k)->Ey;
					dum++;
				}
			}
		}
	}else if(strcmp(field,"Ez")==0){
		for(int i=limx1;i<limx2;i++){
			for(int j=limy1;j<limy2;j++){
				for(int k=limz1;k<limz2;k++){
					dMat[dum/dims[1]+(dum%dims[1])*dims[0]]=ObjectMap3D(i,j,k)->Ez;
					dum++;
				}
			}
		}
	}else if(strcmp(field,"Hx")==0){
		for(int i=limx1;i<limx2;i++){
			for(int j=limy1;j<limy2;j++){
				for(int k=limz1;k<limz2;k++){
					dMat[dum/dims[1]+(dum%dims[1])*dims[0]]=ObjectMap3D(i,j,k)->Hx;
					dum++;
				}
			}
		}
	}else if(strcmp(field,"Hy")==0){
		for(int i=limx1;i<limx2;i++){
			for(int j=limy1;j<limy2;j++){
				for(int k=limz1;k<limz2;k++){
					dMat[dum/dims[1]+(dum%dims[1])*dims[0]]=ObjectMap3D(i,j,k)->Hy;
					dum++;
				}
			}
		}
	}else if(strcmp(field,"Hz")==0){
		for(int i=limx1;i<limx2;i++){
			for(int j=limy1;j<limy2;j++){
				for(int k=limz1;k<limz2;k++){
					dMat[dum/dims[1]+(dum%dims[1])*dims[0]]=ObjectMap3D(i,j,k)->Hz;
					dum++;
				}
			}
		}
	}else{
		printf("Field %s not understood. Exiting...",field);
		exit(EXIT_FAILURE);
	}
	if(dum!=(limx2-limx1)*(limy2-limy1)*(limz2-limz1)){
		printf("Visualization update problem. dum=%d but lims= %d %d %d %d %d %d. Exiting...",dum,limx1,limx2,limy1,limy2,limz1,limz2);
		exit(EXIT_FAILURE);
	}
	
}

void World3D::updateVisualization(){
	copyDataToVisualizer(fieldName,limx1,limx2,limy1,limy2,limz1,limz2);	
	engPutVariable(ep, matlabVarName, matlabField);
	sprintf(buf,"plotField('%s',%d);",matlabVarName,currentTime);
	engEvalString(ep, buf);
}

void World3D::initVisualizer(const char *field,const char *dir, int cutPlane, const char *varName){
	
	limx1 = 0; limx2 = NX; limy1=0; limy2=NY; limz1=0; limz2=NZ;
	strcpy(fieldName,field);
	if(strcmp(dir,"x")==0){
		dims[0]=NY;
		dims[1]=NZ;
		limx1=cutPlane;
		limx2=cutPlane+1;
	}else if(strcmp(dir,"y")==0){
		dims[0]=NX;
		dims[1]=NZ;
		limy1=cutPlane;
		limy2=cutPlane+1;
	}else if(strcmp(dir,"z")==0){
		dims[0]=NX;
		dims[1]=NY;
		limz1=cutPlane;
		limz2=cutPlane+1;
	}else{
		printf("Direction %s not understood. Exiting...",dir);
		exit(EXIT_FAILURE);
	};

	if(ep==0){
		printf("Starting Matlab Engine...\n");
		// Start the matlab engine
		if (!(ep = engOpen("\0"))) {
			printf("Can't start MATLAB engine\n");
			exit(EXIT_FAILURE);
		}
	}
	// Allocate memory for the matlab array
	matlabField = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
	dMat=mxGetPr(matlabField);
	if (dMat == NULL) {
		printf("mxGetPr failed\n");
		exit(EXIT_FAILURE);
	}

	// Initialize the data
	copyDataToVisualizer(fieldName,limx1,limx2,limy1,limy2,limz1,limz2);
	strcpy(matlabVarName,varName);

	// Make the variable global so that the plotField function works
	sprintf(buf,"global %s;",matlabVarName);
	engEvalString(ep, buf);

	// Put the data into Matlab environment and plot it
	engPutVariable(ep, matlabVarName, matlabField);
	
	//  Plot the field
	sprintf(buf,"plotField('%s',%d);",matlabVarName,currentTime);
	engEvalString(ep, buf);
	
}


#endif
