#ifndef World3DGUARD
#define World3DGUARD

// To turn off visualization routines, uncomment the following line
#define NOVISUAL


// Include implemented cell types here
#include "Cell.h"
#include "DielectricECell.h"
#include "PECBrick.h"
#include "PMLCell.h"
#include "LorentzCell.h"
#include "TEMBrick.h"


// Include for printf() function
#include <stdio.h>

// Include for exit() function
#include <stdlib.h>

// For string comparison
#include <string.h>

// Below includes are for reading the MAT files from Matlab
#include <stdio.h>
#include <stdlib.h>
#include "mat.h"

#ifndef NOVISUAL
// For array visualization using Matlab engine
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "engine.h"
#endif


// Below is a simple script to get the (i,j,k)th term of objectMap
// Because matlab stores multidimensional arrays in column major format, so does objectMap.
// objectMap[k*NX*NY+j*NX+i] gives the cell (i,j,k).
#define ObjectMap3D(I,J,K) objectMap3D[(K)*NX*NY+(J)*NX + (I)]

class World3D{

public:

	/******************************************************************************
	***************************Variable declarations*******************************
	*******************************************************************************/


	// Number of cells in x,y,z dimensions
	int NX,NY,NZ;

	// Number of PMLs in x,y,z directions
	int numPMLx, numPMLy, numPMLz;

	// Space and time discretization
	double delX,delY,delZ,delT;

	// Total time steps and current time step
	int totalTimeSteps;
	int currentTime;

	// Source excitation
	double *fieldSource;

	// This variable holds the pointers to all the cells of the simulation domain
	// Because matlab stores multidimensional arrays in column major format, so does objectMap.
	// objectMap3D[k*NX*NY+j*NX+i] gives the cell (i,j,k).
	Cell **objectMap3D;


	// For TF/SF Injection define a source World2D and injection point from that world
	World3D *sourceTF;

	// The teleportation region limits
	int xminTFSF,xmaxTFSF,yminTFSF,ymaxTFSF,zminTFSF,zmaxTFSF;

	// Shift in the location of teleportation box when transferred to targetWorld
	int x0TFSF, y0TFSF, z0TFSF;

	// Indices of cells that should output their contents
	FILE *fp; // Pointer to the file
	int numOutputCells;
	int *ilist;
	int *jlist;
	int *klist;
	/******************************************************************************
	***************************Function definitions********************************
	*******************************************************************************/

	// Constructor and destructor functions
	World3D(const char *file, const char *varName,const char *fileName=0);
	virtual ~World3D() {};
	
	// Assign the boundary cells' floating neighbours to PECBrick cells
	// Do it for all six faces
	virtual void addBoundaryBricks();
	
	// Make sure everything is set up correctly
	virtual void sanityCheck();
	
	// Read the excitation
	void getSource(MATFile *pmat, const char *sourceName);

	// Propagate fields by one time step
	void propagateE();
	void propagateH();

	// Source the E and H fields by the excitation read from the file
	void TFSourceE();
	void TFSourceH();

	// Sources the E and H fields with current densities to teleport fields from 2D->3D
	void TFSFsourceE(bool isTE);
	void TFSFsourceH(bool isTE);

	// sets TF Source World2D
	void setSourceWorld(World3D *source,int xmin,int xmax,int ymin, int ymax, int zmin,int zmax, int x0, int y0, int z0);

	void outputFields();

#ifndef NOVISUAL
	// Field Visualization variables
	double *dMat;
	mxArray *matlabField;
	char matlabVarName[200];

	// Matlab Engine Pointer
	// It's defined to be static so that there is only one copy of Matlab running at all times
	static Engine *ep;

	// Buffer for sending command to Matlab
	char buf[80];

	// Limits of updates for field visualization
	int limx1,limx2,limy1,limy2,limz1,limz2;
	// Field name to be visualized
	char fieldName[10];
	
	// The dimensions of the array for visualization
	int dims[2];

	// Field visualization routines
	// Updates the visualization arrays
	void copyDataToVisualizer(const char *field,int limx1, int limx2,int limy1,int limy2, int limz1,int limz2);
	// Refreshes the visualization screen
	void updateVisualization();
	// Initializes the visualizer
	void initVisualizer(const char *field,const char *dir, int cutPlane, const char *varName);
#endif

protected:

	/******************************************************************************
	***************************Variable declarations*******************************
	*******************************************************************************/
	// This variable holds the pointers to the computation domain limiting PECBrick objects
	Cell **brickMap;
	int numBricks;


	/******************************************************************************
	***************************Function definitions********************************
	*******************************************************************************/
	// This function gets the scalar (either int or double) from the MAT file
	template<class T> void getScalar(MATFile *pmat, const char* name, T *num);

	// Reads the mat file that holds the simulation parameters and initializes world3D
	void readMAT(const char *file, const char *varName);
	
	// Creates the objectMap3D
	void getObjectMap3D(MATFile *pmat, const char *varName);

	// Set the neighbours of each cell
	void setNeighbours();

	// Set the numPML variables i.e. the number of PML layers in x,y,z directions
	void setnumPML();

	// Reads the mat file for output indices
	void getOutputIndices(const char *file, const char *variableName);
};


#endif